@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: DummyProto.h,v 1.7 2001/10/03 08:13:37 lav Exp $ */
+/* $Id: DummyProto.h,v 1.8 2001/12/05 15:03:41 lav Exp $ */
 
 #ifndef DUMMYPROTO_H
 #define DUMMYPROTO_H
@@ -37,29 +37,20 @@ public:
    int StoreStatus();
    void Reconfig(const char *) {}
 
-   ListInfo *MakeListInfo();
+   ListInfo *MakeListInfo(const char *path);
    DirList *MakeDirList(ArgV *);
 };
 
-class DummyNoProto : public FileAccess
+class DummyNoProto : public DummyProto
 {
    char *proto;
 public:
    DummyNoProto(const char *p);
    ~DummyNoProto();
 
-   int Do();
-   int Done();
    const char *GetProto();
    FileAccess *Clone();
-   int Read(void *buf,int size);
-   int Write(const void *buf,int size);
-   int StoreStatus();
-   void Reconfig(const char *) {}
    const char *StrError(int err);
-
-   ListInfo *MakeListInfo();
-   DirList *MakeDirList(ArgV *);
 };
 
 #endif
