@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FileCopy.h,v 1.54 2003/12/01 07:46:18 lav Exp $ */
+/* $Id: FileCopy.h,v 1.55 2004/10/15 12:51:31 lav Exp $ */
 
 /*
   classes defined here:
@@ -78,7 +78,7 @@ public:
    bool CanSeek0() { return can_seek0; }
    bool CanSeek(off_t p) { return p==0 ? CanSeek0() : CanSeek(); }
    off_t GetSeekPos() { return seek_pos; }
-   virtual void Seek(off_t offs) { seek_pos=offs; Empty(); eof=false; broken=false; }
+   virtual void Seek(off_t offs);
    virtual off_t GetRealPos() { return pos; }
    virtual int Buffered() { return in_buffer; }
    virtual bool IOReady() { return true; }
