@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FileSet.cc,v 1.33 2002/07/31 14:35:50 lav Exp $ */
+/* $Id: FileSet.cc,v 1.46 2005/02/08 11:03:05 lav Exp $ */
 
 #include <config.h>
 #include "FileSet.h"
@@ -42,6 +42,11 @@
 #include "IdNameCache.h"
 #include "PatternSet.h"
 
+#define NO_SIZE	     (-1L)
+#define NO_SIZE_YET  (-2L)
+#define NO_DATE	     ((time_t)-1L)
+#define NO_DATE_YET  ((time_t)-2L)
+
 void  FileInfo::Merge(const FileInfo& f)
 {
    if(strcmp(name,f.name))
@@ -68,8 +73,10 @@ void FileInfo::SetName(const char *n)
 {
    if(n==name)
       return;
+   // in case of n being tail of name, dup it first
+   char *n1=xstrdup(n);
    xfree(name);
-   name=xstrdup(n);
+   name=n1;
    defined|=NAME;
 }
 
@@ -104,9 +111,14 @@ void FileSet::Add(FileInfo *fi)
       delete fi;
       return;
    }
-   files=files_sort=(FileInfo**)xrealloc(files,(++fnum)*sizeof(*files));
-   memmove(files+pos+1, files+pos, sizeof(*files)*(fnum-pos-1));
+   if(fnum==allocated)
+   {
+      allocated=allocated?allocated*2:16;
+      files=files_sort=(FileInfo**)xrealloc(files,allocated*sizeof(*files));
+   }
+   memmove(files+pos+1, files+pos, sizeof(*files)*(fnum-pos));
    files[pos]=fi;
+   fnum++;
 }
 
 void FileSet::Sub(int i)
@@ -152,30 +164,31 @@ FileSet::FileSet(FileSet const *set)
 {
    ind=set->ind;
    fnum=set->fnum;
-   sorted = false;
+   sorted=false;
    if(fnum==0)
       files=files_sort=0;
    else
-      files=files_sort=(FileInfo**)xmalloc(fnum*sizeof(*files));
+      files=files_sort=(FileInfo**)xmalloc((allocated=fnum)*sizeof(*files));
    for(int i=0; i<fnum; i++)
       files[i]=new FileInfo(*(set->files[i]));
 }
 
 static int (*compare)(const char *s1, const char *s2);
+static int rev_cmp;
 
 static int sort_name(const void *s1, const void *s2)
 {
    const FileInfo *p1 = *(const FileInfo **) s1;
    const FileInfo *p2 = *(const FileInfo **) s2;
-   return compare(p1->name, p2->name);
+   return compare(p1->name, p2->name) * rev_cmp;
 }
 
 static int sort_size(const void *s1, const void *s2)
 {
    const FileInfo *p1 = *(const FileInfo **) s1;
    const FileInfo *p2 = *(const FileInfo **) s2;
-   if(p1->size > p2->size) return -1;
-   if(p1->size < p2->size) return 1;
+   if(p1->size > p2->size) return -rev_cmp;
+   if(p1->size < p2->size) return rev_cmp;
    return 0;
 }
 
@@ -183,8 +196,8 @@ static int sort_dirs(const void *s1, const void *s2)
 {
    const FileInfo *p1 = *(const FileInfo **) s1;
    const FileInfo *p2 = *(const FileInfo **) s2;
-   if(p1->filetype == FileInfo::DIRECTORY && !p2->filetype == FileInfo::DIRECTORY) return -1;
-   if(!p1->filetype == FileInfo::DIRECTORY && p2->filetype == FileInfo::DIRECTORY) return 1;
+   if(p1->filetype == FileInfo::DIRECTORY && !p2->filetype == FileInfo::DIRECTORY) return -rev_cmp;
+   if(!p1->filetype == FileInfo::DIRECTORY && p2->filetype == FileInfo::DIRECTORY) return rev_cmp;
    return 0;
 }
 
@@ -194,15 +207,24 @@ static int sort_rank(const void *s1, const void *s2)
    const FileInfo *p2 = *(const FileInfo **) s2;
    if(p1->GetRank()==p2->GetRank())
       return sort_name(s1,s2);
-   return p1->GetRank()<p2->GetRank() ? -1 : 1;
+   return p1->GetRank()<p2->GetRank() ? -rev_cmp : rev_cmp;
+}
+
+static int sort_date(const void *s1, const void *s2)
+{
+   const FileInfo *p1 = *(const FileInfo **) s1;
+   const FileInfo *p2 = *(const FileInfo **) s2;
+   if(p1->date==p2->date)
+      return sort_name(s1,s2);
+   return p1->date>p2->date ? -rev_cmp : rev_cmp;
 }
 
 /* files_sort is an alias of files when sort == NAME (since
  * files is always sorted by name), and an independant array
  * of pointers (pointing to the same data) otherwise. */
-void FileSet::Sort(sort_e newsort, bool casefold)
+void FileSet::Sort(sort_e newsort, bool casefold, bool reverse)
 {
-   if(newsort == BYNAME && !casefold) {
+   if(newsort == BYNAME && !casefold && !reverse) {
       Unsort();
       return;
    }
@@ -213,16 +235,19 @@ void FileSet::Sort(sort_e newsort, bool casefold)
 	 files_sort[i] = files[i];
    }
 
-   sorted = true;
+   sorted=true;
 
    if(casefold) compare = strcasecmp;
    else compare = strcmp;
 
+   rev_cmp=(reverse?-1:1);
+
    switch(newsort) {
    case BYNAME: qsort(files_sort, fnum, sizeof(FileInfo *), sort_name); break;
    case BYSIZE: qsort(files_sort, fnum, sizeof(FileInfo *), sort_size); break;
    case DIRSFIRST: qsort(files_sort, fnum, sizeof(FileInfo *), sort_dirs); break;
    case BYRANK: qsort(files_sort, fnum, sizeof(FileInfo *), sort_rank); break;
+   case BYDATE: qsort(files_sort, fnum, sizeof(FileInfo *), sort_date); break;
    }
 }
 
@@ -232,8 +257,8 @@ void FileSet::Unsort()
 {
    if(!sorted) return;
    xfree(files_sort);
-   files_sort = files;
-   sorted = false;
+   files_sort=files;
+   sorted=false;
 }
 
 void FileSet::Empty()
@@ -244,6 +269,7 @@ void FileSet::Empty()
    xfree(files);
    files=0; files_sort=0;
    fnum=0;
+   allocated=0;
    ind=0;
 }
 
@@ -275,6 +301,16 @@ void FileSet::SubtractNotIn(const FileSet *set)
       if(!set->FindByName(files[i]->name))
 	 Sub(i--);
 }
+void FileSet::SubtractSameType(const FileSet *set)
+{
+   for(int i=0; i<fnum; i++)
+   {
+      FileInfo *f=set->FindByName(files[i]->name);
+      if(f && files[i]->defined&FileInfo::TYPE && f->defined&FileInfo::TYPE
+      && files[i]->filetype==f->filetype)
+	 Sub(i--);
+   }
+}
 
 void FileSet::SubtractOlderThan(time_t t)
 {
@@ -302,6 +338,18 @@ void FileSet::SubtractDirs()
       }
    }
 }
+void FileSet::SubtractNotDirs()
+{
+   for(int i=0; i<fnum; i++)
+   {
+      if(!(files[i]->defined&FileInfo::TYPE)
+      || files[i]->filetype!=FileInfo::DIRECTORY)
+      {
+	 Sub(i);
+	 i--;
+      }
+   }
+}
 
 void FileSet::ExcludeDots()
 {
@@ -337,7 +385,7 @@ bool  FileInfo::SameAs(const FileInfo *fi,int ignore)
       if(p<fi->date_prec)
 	 p=fi->date_prec;
       if(!(ignore&IGNORE_DATE_IF_OLDER && date<fi->date)
-      && labs((long)date-(long)(fi->date))>p)
+      && labs(date-fi->date)>p)
 	 return false;
    }
 
@@ -361,11 +409,7 @@ void FileSet::Count(int *d,int *f,int *s,int *o)
 {
    for(int i=0; i<fnum; i++)
    {
-      if(!(files[i]->defined&FileInfo::TYPE))
-      {
-	 if(o) (*o)++;
-      }
-      else switch(files[i]->filetype)
+      switch(files[i]->filetype)
       {
       case(FileInfo::DIRECTORY):
 	 if(d) (*d)++; break;
@@ -373,6 +417,8 @@ void FileSet::Count(int *d,int *f,int *s,int *o)
 	 if(f) (*f)++; break;
       case(FileInfo::SYMLINK):
 	 if(s) (*s)++; break;
+      case(FileInfo::UNKNOWN):
+	 if(o) (*o)++;
       }
    }
 }
@@ -516,7 +562,7 @@ void FileSet::LocalChmod(const char *dir,mode_t mask)
 	 struct stat st;
 	 mode_t new_mode=file->mode&~mask;
 
-	 if(stat(local_name,&st)!=-1 && st.st_mode!=new_mode)
+	 if(stat(local_name,&st)!=-1 && (st.st_mode&07777)!=new_mode)
 	    chmod(local_name,new_mode);
       }
    }
@@ -589,12 +635,19 @@ void FileSet::SubtractCurr()
 
 void FileInfo::Init()
 {
+   filetype=UNKNOWN;
+   mode=(mode_t)-1;
+   date=NO_DATE;
+   date_prec=0;
+   size=NO_SIZE;
+   nlinks=0;
    name=0;
    defined=0;
    symlink=0;
    data=0;
    user=0; group=0;
    rank=0;
+   longname=0;
 }
 
 FileInfo::FileInfo(const FileInfo &fi)
@@ -611,6 +664,7 @@ FileInfo::FileInfo(const FileInfo &fi)
    date_prec=fi.date_prec;
    size=fi.size;
    nlinks=fi.nlinks;
+   longname=xstrdup(fi.longname);
 }
 
 #ifndef S_ISLNK
@@ -670,11 +724,167 @@ check_again:
 #endif /* HAVE_LSTAT */
 }
 
+/* parse_ls_line: too common procedure to make it protocol specific */
+/*
+-rwxr-xr-x   1 lav      root         4771 Sep 12  1996 install-sh
+-rw-r--r--   1 lav      root         1349 Feb  2 14:10 lftp.lsm
+drwxr-xr-x   4 lav      root         1024 Feb 22 15:32 lib
+lrwxrwxrwx   1 lav      root           33 Feb 14 17:45 ltconfig -> /usr/share/libtool/ltconfig
+NOTE: group may be missing.
+*/
+FileInfo *FileInfo::parse_ls_line(const char *line_c,const char *tz)
+{
+   char *line=alloca_strdup(line_c);
+   char *next=0;
+   FileInfo *fi=0; /* don't instantiate until we at least have something */
+#define FIRST_TOKEN strtok_r(line," \t",&next)
+#define NEXT_TOKEN  strtok_r(NULL," \t",&next)
+#define ERR do{delete fi;return(0);}while(0)
+
+   /* parse perms */
+   char *t = FIRST_TOKEN;
+   if(t==0)
+      ERR;
+
+   fi = new FileInfo;
+   switch(t[0])
+   {
+   case('l'):  // symlink
+      fi->SetType(fi->SYMLINK);
+      break;
+   case('d'):  // directory
+      fi->SetType(fi->DIRECTORY);
+      break;
+   case('-'):  // plain file
+      fi->SetType(fi->NORMAL);
+      break;
+   case('b'): // block
+   case('c'): // char
+   case('p'): // pipe
+   case('s'): // sock
+   case('D'): // Door
+      return 0;  // ignore
+   default:
+      ERR;
+   }
+   mode_t mode=parse_perms(t+1);
+   if(mode!=(mode_t)-1)
+      fi->SetMode(mode);
+
+   // link count
+   t = NEXT_TOKEN;
+   if(!t)
+      ERR;
+   fi->SetNlink(atoi(t));
+
+   // user
+   t = NEXT_TOKEN;
+   if(!t)
+      ERR;
+   fi->SetUser(t);
+
+   // group or size
+   char *group_or_size = NEXT_TOKEN;
+
+   // size or month
+   t = NEXT_TOKEN;
+   if(!t)
+      ERR;
+   if(isdigit(*t))
+   {
+      // it's size, so the previous was group:
+      fi->SetGroup(group_or_size);
+      long long size;
+      if(sscanf(t,"%lld",&size)==1)
+	 fi->SetSize(size);
+      t = NEXT_TOKEN;
+      if(!t)
+	 ERR;
+   }
+   else
+   {
+      // it was month, so the previous was size:
+      long long size;
+      if(sscanf(group_or_size,"%lld",&size)==1)
+	 fi->SetSize(size);
+   }
+
+   struct tm date;
+   memset(&date,0,sizeof(date));
+
+   date.tm_mon=parse_month(t);
+   if(date.tm_mon==-1)
+      date.tm_mon=0;
+
+   const char *day_of_month = NEXT_TOKEN;
+   if(!day_of_month)
+      ERR;
+   date.tm_mday=atoi(day_of_month);
+
+   bool year_anomaly=false;
+
+   // time or year
+   t = NEXT_TOKEN;
+   if(!t)
+      ERR;
+   date.tm_hour=date.tm_min=0;
+   int prec=30;
+   if(strlen(t)==5)
+   {
+      sscanf(t,"%2d:%2d",&date.tm_hour,&date.tm_min);
+      date.tm_year=guess_year(date.tm_mon,date.tm_mday,date.tm_hour,date.tm_min) - 1900;
+   }
+   else
+   {
+      if(day_of_month+strlen(day_of_month)+1 == t)
+	 year_anomaly=true;
+      date.tm_year=atoi(t)-1900;
+      /* We don't know the hour.  Set it to something other than 0, or
+       * DST -1 will end up changing the date. */
+      date.tm_hour = 12;
+      prec=12*60*60;
+   }
+
+   date.tm_isdst=-1;
+   date.tm_sec=30;
+
+   fi->SetDate(mktime_from_tz(&date,tz),prec);
+
+   char *name=strtok_r(NULL,"",&next);
+   if(!name)
+      ERR;
+
+   // there are ls which outputs extra space after year.
+   if(year_anomaly && *name==' ')
+      name++;
+
+   if(fi->filetype==fi->SYMLINK)
+   {
+      char *arrow=name;
+      while((arrow=strstr(arrow," -> "))!=0)
+      {
+	 if(arrow!=name && arrow[4]!=0)
+	 {
+	    *arrow=0;
+	    fi->SetSymlink(arrow+4);
+	    break;
+	 }
+	 arrow++;
+      }
+   }
+   fi->SetName(name);
+   fi->SetLongName(line_c);
+
+   return fi;
+}
+
+
 FileInfo::~FileInfo()
 {
    xfree(name);
    xfree(symlink);
    xfree(data);
+   xfree(longname);
 }
 
 int FileSet::Have() const
@@ -711,3 +921,51 @@ void FileSet::SortByPatternList(const char *list_c)
 	    files[i]->SetRank(rank);
    Sort(BYRANK);
 }
+
+void FileInfo::MakeLongName()
+{
+   longname=(char*)xrealloc(longname,80+xstrlen(name)+xstrlen(symlink));
+   char filetype_c='-';
+   switch(filetype)
+   {
+   case NORMAL:	   break;
+   case UNKNOWN:   break;
+   case DIRECTORY: filetype_c='d'; break;
+   case SYMLINK:   filetype_c='l'; break;
+   }
+   int mode1=(defined&MODE?mode:
+      (filetype_c=='d'?0755:(filetype_c=='l'?0777:0644)));
+   sprintf(longname,"%c%s  ",filetype_c,format_perms(mode1));
+   char usergroup[33];
+   usergroup[0]=0;
+   if(defined&(USER|GROUP))
+      sprintf(usergroup,"%.16s%s%.16s",defined&USER?user:"?",
+		  defined&GROUP?"/":"",defined&GROUP?group:"");
+   char size_str[20];
+   strcpy(size_str,"-");
+   if(defined&SIZE)
+      sprintf(size_str,"%lld",(long long)size);
+   int w=20-strlen(usergroup);
+   if(w<1)
+      w=1;
+   sprintf(longname+strlen(longname),"%s %*s ",usergroup,w,size_str);
+   const char *date_str="-";
+   if(defined&DATE)
+      date_str=TimeDate(date).IsoDateTime();
+   sprintf(longname+strlen(longname),"%s %s",date_str,name);
+   if(defined&SYMLINK_DEF)
+      sprintf(longname+strlen(longname)," -> %s",symlink);
+}
+
+int FileSet::EstimateMemory() const
+{
+   int size=sizeof(FileSet)+sizeof(FileInfo*)*allocated;
+   for(int i=0; i<fnum; i++)
+   {
+      size+=sizeof(FileInfo);
+      size+=xstrlen(files[i]->name);
+      size+=xstrlen(files[i]->symlink);
+      size+=xstrlen(files[i]->longname);
+   }
+   return size;
+}
