@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: Filter.h,v 1.13 2001/10/05 06:26:51 lav Exp $ */
+/* $Id: Filter.h,v 1.18 2005/03/13 12:17:59 lav Exp $ */
 
 #ifndef FILTER_H
 #define FILTER_H
@@ -29,10 +29,13 @@ class ArgV;
 
 class FDStream
 {
+   bool close_fd;
+
 public:
    int fd;
    char *name;
    char *error_text;
+   const char *status;
 
    bool	 error() { return error_text!=0; }
 
@@ -42,7 +45,11 @@ public:
    FDStream(int new_fd,const char *new_name);
    virtual ~FDStream();
 
-   void MakeErrorText();
+   void MakeErrorText(int e=0);
+   bool NonFatalError(int err);
+   void set_status(const char *str) { status=str; }
+   void clear_status() { status=0; }
+   void CloseFD() { close_fd=true; }
 
    virtual off_t get_size() { return -1; }
    virtual void setmtime(time_t) {}
@@ -63,6 +70,7 @@ class OutputFilter : public FDStream
    ProcWait *w;
    pid_t pg;
    FDStream *second;
+   bool delete_second;
    bool stderr_to_stdout;
 
    char *oldcwd;
@@ -90,11 +98,13 @@ public:
    int getfd();
    bool Done();
 
-   bool usesfd(int n_fd) { return FDStream::usesfd(n_fd) || n_fd<=2; }
-   void Kill(int sig=SIGTERM) { if(w) w->Kill(sig); }
+   bool usesfd(int n_fd);
+   void Kill(int sig=SIGTERM);
    pid_t GetProcGroup() { return pg; }
 
    bool broken();
+
+   void DeleteSecondaryStream() { delete_second=true; }
 };
 
 class InputFilter : public OutputFilter
