@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: NetAccess.h,v 1.29 2002/07/18 09:13:49 lav Exp $ */
+/* $Id: NetAccess.h,v 1.36 2005/03/01 13:53:11 lav Exp $ */
 
 #ifndef NETACCESS_H
 #define NETACCESS_H
@@ -39,6 +39,7 @@ protected:
 
    int	 max_retries;
    int	 max_persist_retries;
+   int	 retries;
    int	 persist_retries;
 
    int	 idle;
@@ -48,32 +49,43 @@ protected:
    bool	 CheckTimeout();
 
    int	 reconnect_interval;
+   float reconnect_interval_current;
    float reconnect_interval_multiplier;
    int   reconnect_interval_max;
 
    int	 connection_limit;
    bool	 connection_takeover;
 
+   class RateLimit *rate_limit;
+
    int	 socket_buffer;
    int	 socket_maxseg;
    void	 SetSocketBuffer(int sock);
    void	 SetSocketMaxseg(int sock);
+
    static void KeepAlive(int sock);
+   static void MinimizeLatency(int sock);
+   static void MaximizeThroughput(int sock);
    static void ReuseAddress(int sock);
    static void SetSocketBuffer(int sock,int val);
    static void SetSocketMaxseg(int sock,int val);
-
-   class RateLimit *rate_limit;
-
+   static int SocketBuffered(int sock);
    static const char *SocketNumericAddress(const sockaddr_u *u);
    static int SocketPort(const sockaddr_u *u);
    static socklen_t SocketAddrLen(const sockaddr_u *u);
    static int SocketConnect(int fd,const sockaddr_u *u);
+   int SocketCreate(int,int,int);
+   int SocketCreateTCP(int);
 
    char	 *proxy;
    char	 *proxy_port;
    char  *proxy_user;
    char  *proxy_pass;
+   char  *proxy_proto;
+
+   char	 *home_auto;
+   void	 PropagateHomeAuto();
+   const char *FindHomeAuto();
 
    void SayConnectingTo();
 
@@ -82,10 +94,10 @@ protected:
 
    int Resolve(const char *defp,const char *ser,const char *pr);
 
-   long ReconnectInterval();
    const char *DelayingMessage();
    bool ReconnectAllowed();
-   bool NextTry();   // returns false if max-retries exceeded.
+   bool NextTry();	// returns false if max-retries exceeded.
+   void TrySuccess();	// reset retry counters.
 
    virtual void HandleTimeout();
 
@@ -155,7 +167,7 @@ protected:
    int get_info_cnt;
 
    int mode;
-   Buffer *ubuf;
+   IOBuffer *ubuf;
 
    bool get_time_for_dirs;
    bool can_get_prec_time;
